;;; erlang.ss --- Scheme interface to Erlang.
;;; Copyleft (]) 2008 Eric Sessoms <nubgames@gmail.com>

;; based on distel http://code.google.com/p/distel
;; Copyleft (]) 2000-2002 Luke Gorrie <luke@bluetail.com>

;; originally from Lennart Staflin's elisp CORBA client

#lang scheme

(require scheme/port
         (planet "digest.ss" ("soegaard" "digest.plt" 1 2))
         "pack.ss")

(define erlang-node-name 'scheme@localhost)

(define (erlang-set-node-name! node-name)
  (set! erlang-node-name node-name))

(define (symbol->bytes sym)
  (string->bytes/latin-1 (symbol->string sym)))

;; Erlang Port Mapper Daemon
(define epmd-port 4369)
(define epmd-host "localhost")

(define (erlang-node->parts node-name)
  (match (regexp-split #px"@" (symbol->string node-name))
    [(list name host) (values name host)]
    [(list name)      (values name epmd-host)]))

(define (epmd-write-message msg out)
  (pack "n/a" #:out out msg))

(define (epmd-with-message msg resp #:host host)
  (let-values ([(in out) (tcp-connect (or host epmd-host) epmd-port)])
    (epmd-write-message msg out)
    (close-output-port out)
    (let ([result (resp in)])
      (close-input-port in)
      result)))

(define (epmd-get-names (host #f))
  (epmd-with-message #"n" #:host host
    (lambda (in)
      (let* ([msg   (unpack "N/A" in)]
             [names (regexp-split #px"\n" msg)])
        (reverse (cdr (reverse names)))))))

(define (epmd-get-port node-name)
  (let-values ([(name host) (erlang-node->parts node-name)])
    (epmd-with-message (pack "aA" #"p" name) #:host host
      (lambda (in)
        (unpack "n" in)))))

;; Returns port, node-type, protocol, vsn-max, vsn-min, name, extra.
(define (epmd-get-port-r4 node-name)
  (let-values ([(name host) (erlang-node->parts node-name)])
    (epmd-with-message (pack "aA" #"z" name) #:host host
      (lambda (in)
        (let ([result (read-bytes 2 in)])
          (when (bytes=? #"w\0" result)
                (unpack-values "nCCnnn/An/A" in)))))))

;; R4 and later.
(define erlang-version 5)

;; Optional feature flags
(define flag-published            #x01)
(define flag-atom-cache           #x02)
(define flag-extended-references  #x04)
(define flag-dist-monitor         #x08)
(define flag-fun-tags             #x10)
(define flag-dist-monitor-name    #x20)
(define flag-hidden-atom-cache    #x40)
(define flag-new-fun-tags         #x80)
(define flag-extended-pids-ports #x100)

(define (send-message out #:hdrlen (hdrlen 4) . msg)
  (pack (if (= hdrlen 4) "N/a" "n/a") #:out out (apply bytes-append msg))
  (flush-output out))

(define (take-message in #:hdrlen (hdrlen 4))
  "Read a length-prefixed message from an input port."
  (let ([msg (unpack (if (= hdrlen 4) "N/a" "n/a") in)])
    msg))

(define (get-cookie (cookie #f))
  "Read the cookie from the default location, as necessary."
  (if cookie
      cookie
      (call-with-input-file* (expand-user-path "~/.erlang.cookie") #:mode 'text
        read-bytes-line)))

(define (erlang-digest challenge (cookie #f))
  "Find the MD5 digest of the concatenation of the cookie and the challenge."
  (let* ([digest
          (md5 (bytes-append (get-cookie cookie)
                             (string->bytes/latin-1 (number->string challenge))))]
         [value (string->number digest 16)])
    (integer->network value 16)))

;; Implements the Erlang distribution protocol handshake.
(define (handshake in out (cookie #f))
  (define (send-name)
    (send-message out #:hdrlen 2
                  (pack "anNa" #"n" erlang-version
                        (bitwise-ior flag-extended-references
                                     flag-extended-pids-ports)
                        (symbol->bytes erlang-node-name))))
  (define (recv-status)
    (let ([msg (take-message in #:hdrlen 2)])
      (bytes=? msg #"sok")
      msg))
  (define (recv-challenge)
    (let*-values ([(msg)                         (take-message in #:hdrlen 2)]
                  [(tag version flags challenge) (unpack-values "CnNN" msg)])
      challenge))
  (define (challenge-reply digest)
    (send-message out #:hdrlen 2
                  #"r" (integer->integer-bytes 42 4 #f #t)
                  digest))
  (define (recv-challenge-ack)
    (let* ([msg (take-message in #:hdrlen 2)]
           [reply (subbytes msg 1 17)])
      (bytes=? reply (erlang-digest 42 cookie))))

  ;; Start here.
  (send-name)
  (recv-status)
  (let ([challenge (recv-challenge)])
    (challenge-reply (erlang-digest challenge cookie)))
  (recv-challenge-ack))

(define (erlang-connect node-name #:cookie (cookie #f))
  (let*-values ([(name host) (erlang-node->parts node-name)]
                [(in out)    (tcp-connect host (epmd-get-port node-name))])
    (handshake in out cookie)
    (values in out)))

;; -----------------------------------------------------------------------------
;; Library for encoding/decoding elisp terms into erlang's external
;; term format.  For format details see erts/emulator/internal_doc/ in
;; the Erlang/OTP sources.
;;
;; Supported mappings from/to erlext to elisp:
;;   atom    -> symbol
;;   string  -> string
;;   integer -> integer
;;   list    -> list
;;   tuple   -> #(...)
;;   pid     -> #(ERL-TAG 'pid node id serial creation)
;;   binary  -> bytes
;; Not mapped/supported yet:
;;   ref, port, float, bignum, function, ...

;; Type tags.
(define erlang-tag-alist
  '((smallInt	. 97)     ; a
    (int	. 98)     ; b
    (float	. 99)     ; c
    (atom	. 100)    ; d
    (cached	. 67)     ; C
    (ref	. 101)    ; e
    (port	. 102)    ; f
    (pid	. 103)    ; g
    (smallTuple . 104)    ; h
    (largeTuple . 105)    ; i
    (null	. 106)    ; j
    (string	. 107)    ; k
    (list	. 108)    ; l
    (bin	. 109)    ; m
    (smallBig	. 110)    ; n
    (largeBig	. 111)    ; o
    (newRef	. 114)))  ; r

(define (assr v lst)
  "Locates the first element of lst whose cdr is eq? to v.  Returns the pair or #f."
  (let ([result (memf (lambda (x) (eq? v (cdr x))) lst)])
    (if result
        (car result)
        #f)))

;; smallInt -> 97
(define (erlang-get-code tag)
  (cdr (assq tag erlang-tag-alist)))

;; 97 -> smallInt
(define (erlang-get-tag number)
  (car (assr number erlang-tag-alist)))

;; Tag placed in the first element of a vector to indicate a non-tuple type.
(define erlang-tag 'TYPE)

(define (erlang-read-object in)
  (define (read-string)
    (unpack "n/A" in))
  (define (read-atom)
    (string->symbol (read-string)))
  (define (read-tuple arity)
    (apply vector (for/list ([i (in-range arity)]) (erlang-read-object in))))
  (define (read-small-tuple)
    (read-tuple (unpack "C" in)))
  (define (read-large-tuple)
    (read-tuple (unpack "N" in)))
  (define (read-list)
    (let* ([arity  (unpack "N" in)]
           [result (for/list ([i (in-range arity)])
                     (erlang-read-object in))])
      ;; We find nulls at the end of lists, not counted in arity.
      (eq? (erlang-get-code 'null) (unpack "C" in))
      result))
  (define (read-binary)
    (unpack "N/a" in))
  (define (read-pid)
    (apply vector erlang-tag 'erlang-pid
           (erlang-read-object in)  ; node
           (unpack "NNC" in)))      ; id serial creation
  (define (read-port)
    (apply vector erlang-tag 'erlang-port
           (erlang-read-object in)  ; node
           (unpack "NC" in)))       ; id creation
  (define (read-ref)
    (apply vector erlang-tag 'erlang-ref
           (erlang-read-object in)  ; node
           (unpack "NC" in)))       ; id creation
  (define (read-newref)
    (let ([len (unpack "n" in)])
      (vector erlang-tag 'erlang-new-ref
              (erlang-read-object in)       ; node
              (unpack "C" in)               ; creation
              (read-bytes (* 4 len) in))))  ; id
  (define (read-small-bignum)
    (let ([len (unpack "C" in)])
      (network->integer (read-bytes len in))))
  (define (read-large-bignum)
    (let ([len (unpack "N" in)])
      (network->integer (read-bytes len in))))

  ;; Start here.
  (case (erlang-get-tag (unpack "C" in))
    [(smallInt)   (unpack "C" in)]
    [(int)        (unpack "N" in)]
    [(atom)       (read-atom)]
    [(smallTuple) (read-small-tuple)]
    [(largeTuple) (read-large-tuple)]
    [(list)       (read-list)]
    [(string)     (read-string)]
    [(bin)        (read-binary)]
    [(null)       empty]
    [(pid)        (read-pid)]
    [(port)       (read-port)]
    [(ref)        (read-ref)]
    [(newRef)     (read-newref)]
    [(smallBig)   (read-small-bignum)]
    [(largeBig)   (read-large-bignum)]
    [else         (error "unknown tag")]))

(define erlang-protocol-version 131)

(define (with-input-from in f)
  (if (bytes? in)
      (let* ([input  (open-input-bytes in)]
             [result (f input)])
        (close-input-port input)
        result)
      (f in)))

(define (erlang-binary->term in)
  (with-input-from in
    (lambda (input)
      (eq? (unpack "C" input) erlang-protocol-version)
      (erlang-read-object input))))

(define (erlang-write-object obj out)
  (define (write-null)
    (pack "C" (erlang-get-code 'null) #:out out))
  (define (write-list lst)
    (if (null? lst)
        (write-null)
        (begin
          (pack "CN" #:out out (erlang-get-code 'list) (length lst))
          (for ([term (in-list lst)])
            (erlang-write-object term out))
          (write-null))))
  (define (write-string str)
    (pack "Cn/A" #:out out (erlang-get-code 'string) str))
  (define (write-atom atom)
    (pack "Cn/A" #:out out (erlang-get-code 'atom) (symbol->string atom)))
  (define (write-binary bstr)
    (pack "CN/a" #:out out (erlang-get-code 'bin) bstr))
  (define (write-int n)
    (cond
     [(< n 256) (pack "CC" #:out out (erlang-get-code 'smallInt) n)]
     [#t        (pack "CN" #:out out (erlang-get-code 'int)      n)]))
  (define (write-tuple tup)
    (define (write-pid node id serial creation)
      (pack "C" #:out out (erlang-get-code 'pid))
      (erlang-write-object node out)
      (pack "NNC" #:out out id serial creation))
    (define (write-port node id creation)
      (pack "C" #:out out (erlang-get-code 'port))
      (erlang-write-object node out)
      (pack "NC" #:out out id creation))
    (define (write-ref node id creation)
      (pack "C" #:out out (erlang-get-code 'ref))
      (erlang-write-object node out)
      (pack "NC" #:out out id creation))
    (define (write-new-ref node creation id)
      (pack "Cn" #:out out (erlang-get-code 'newRef) (/ (bytes-length id) 4))
      (erlang-write-object node out)
      (pack "C" #:out out creation)
      (write-bytes id out))
    (let ([arity (vector-length tup)])
      (if (and (> arity 1) (eq? erlang-tag (vector-ref tup 0)))
          (let ([f (case (vector-ref tup 1)
                     [(erlang-pid)     write-pid]
                     [(erlang-port)    write-port]
                     [(erlang-ref)     write-ref]
                     [(erlang-new-ref) write-new-ref])])
            (apply f (cddr (vector->list tup))))
          (begin
            (if (< arity 256)
                (pack "CC" #:out out (erlang-get-code 'smallTuple) arity)
                (pack "CN" #:out out (erlang-get-code 'largeTuple) arity))
            (for ([term (in-vector tup)])
              (erlang-write-object term out))))))

  ;; Start here.
  (cond
   [(bytes? obj)   (write-binary obj)]
   [(integer? obj) (write-int obj)]
   [(list? obj)    (write-list obj)]
   [(string? obj)  (write-string obj)]
   [(symbol? obj)  (write-atom obj)]
   [(vector? obj)  (write-tuple obj)]))

(define (with-output-to out f)
  (if (not out)
      (let ([output (open-output-bytes)])
        (f output)
        (close-output-port output)
        (get-output-bytes output))
      (f out)))

(define (term->erlang-binary term (out #f))
  (with-output-to out
    (lambda (output)
      (pack "C" erlang-protocol-version #:out output)
      (erlang-write-object term output))))

(define (erlang-receive in (len #f))
  (let ([bstr (if len
                  (read-bytes len in)
                  (take-message in))])
    (if (zero? (bytes-length bstr))
        (values 'keep-alive null)
        (let* ([bytes (open-input-bytes bstr)]
               [type  (unpack "C" bytes)]
               [ctl   (erlang-binary->term bytes)]
               [msg   (case (vector-ref ctl 0)
                        [(2 6) (erlang-binary->term bytes)]
                        [else  #f])])
          (close-input-port bytes)
          (eq? ctl 112)  ; pass-through
          (values ctl msg)))))

(define (erlang-send-request out ctl (msg #f))
  (apply send-message out
         (pack "C" 121)  ; pass-through
         (term->erlang-binary ctl)
         (if msg
             (list (term->erlang-binary msg))
             empty)))

(define empty-symbol (string->symbol ""))

(define (erl-send out pid msg)
  (erlang-send-request out (vector 2 empty-symbol pid) msg))

(define (erlang-send node pid msg)
  (node erl-send pid msg))

(define (erl-reg-send out from to term)
  (erlang-send-request out (vector 6 from empty-symbol to) term))

(define (erlang-reg-send node from to term)
  (node erl-reg-send from to term))

(define (erlang-link out from to)
  (erlang-send-request out (vector 1 from to)))

(define (erlang-unlink out from to)
  (erlang-send-request out (vector 4 from to)))

(define (erlang-exit out from to reason)
  (erlang-send-request out (vector 3 from to reason)))

(define (erlang-exit2 out from to reason)
  (erlang-send-request out (vector 8 from to reason)))

;;------------------------------------------------------------------------------
;; Process Management
;;------------------------------------------------------------------------------

;; Generate a unique object reference.
(define erlang-gen-ref
  (let ([id 0])
    (lambda ()
      (set! id (add1 id))
      (vector erlang-tag 'erlang-new-ref erlang-node-name 1 (pack "NNN" id 0 0)))))

;; Generate a unique process identifier.
(define erlang-gen-pid
  (let ([id 0])
    (lambda ()
      (set! id (add1 id))
      (vector erlang-tag 'erlang-pid erlang-node-name id 0 1))))

(define (erlang-pid? obj)
  "Does 'obj' represent an Erlang process identifier?"
  (and (vector? obj)
       (eq? (vector-length obj) 7)
       (eq? (vector-ref obj 0) erlang-tag)
       (eq? (vector-ref obj 1) 'erlang-pid)))

(define (erlang-pid->id pid)
  "Extract the id from a pid, because that's all we usually need internally."
  (vector-ref pid 3))

(define (erlang-ref pid)
  "We can refer to processes either by pid or id."
  (if (erlang-pid? pid)
      (erlang-pid->id pid)
      pid))

;; Global hash table containing references to all our local processes.
(define erl-procs (make-hash))

(define (erlang-find-proc pid)
  "Find a local process by pid or id."
  (hash-ref erl-procs (erlang-ref pid) #f))

(define (erlang-send-proc pid . args)
  "Send a message to the identified process."
  (let ([func (erlang-find-proc pid)])
    (if func
        (apply func args)
        #f)))

(define (erlang-register pid func)
  "Globally register a process so that we can find it later."
  (if (erlang-find-proc pid)
      #f
      (hash-set! erl-procs (erlang-ref pid) func)))

(define (erlang-unregister name)
  "Unregister a process."
  (hash-remove! erl-procs name))

;; Thread local---every process needs an easy way to refer to itself.
(define erlang-self (make-parameter null))

(define (erlang-spawn thunk #:reg (reg #f))
  "Start a new process/thread, optionally registering it under a well-known name."
  (parameterize ([erlang-self (erlang-gen-pid)])
    (let* ([thd  (thread thunk)]
           [func (case-lambda
                    [(arg) (thread-send thd arg)]
                    [args  (thread-send thd args)])])
      (erlang-register (erlang-self) func)
      (when reg
            (erlang-register reg func))
      func)))

(define (erlang-console)
  "Setup to recieve messages for the console, for interactive development."
  (erlang-self (erlang-gen-pid))
  (let* ([self (current-thread)]
         [func (case-lambda
                  [(arg) (thread-send self arg)]
                  [args  (thread-send self args)])])
    (erlang-register (erlang-self) func)
    (erlang-register 'console func)))

(define (erl-dispatch ctl msg)
  "Use the ctl to address the msg to its destination."
  (let ([pid (case (vector-ref ctl 0)
              [(2) (vector-ref ctl 2)]  ; send:     identified by pid
              [(6) (vector-ref ctl 3)]  ; reg-send: identified by name
              [else #f])])
    (when pid
          (erlang-send-proc pid msg))))

(define ((erl-node-proc node-name))
  "Manages communication with a remote Erlang node."
  (let-values ([(in out) (erlang-connect node-name)])
    (letrec ([loop
              (lambda ()
                (let ([result (sync (read-bytes-evt 4 in) (thread-receive-evt))])
                  (if (bytes? result)
                      ;; If we receive data over the wire, parse it and act on it.
                      (let-values ([(ctl msg) (erlang-receive in (unpack "N" result))])
                        (cond
                         [(eq? ctl 'keep-alive) (send-message out #"")]
                         [msg                   (erl-dispatch ctl msg)]
                         [#t                    #f]))
                      ;; If we receive a message fron another thread, send it on.
                      (let ([msg (thread-receive)])
                        (apply (car msg) out (cdr msg)))))
                (loop))])
      (loop))))

(define (erlang-node node-name)
  "Returns a function that stands-in as a proxy for an Erlang node."
  (let ([proc (erlang-find-proc node-name)])
    (if proc
        proc
        (erlang-spawn (erl-node-proc node-name) #:reg node-name))))

;;------------------------------------------------------------------------------
;; High-Level Communication
;;------------------------------------------------------------------------------

(define (erlang-gen-call node pid m f . a)
  "A call (two-way request/response) using the gen_server protocol."
  (erlang-reg-send node (erlang-self) pid
                   (vector '$gen_call
                           (vector (erlang-self) (erlang-gen-ref))
                           (vector 'call m f a (erlang-self))))
  ;; Send message and wait for reply, pretending its synchronous.
  (thread-receive))

(define (erlang-gen-cast node pid m f . a)
  "A cast (one-way notification) using the gen_server protocol."
  (erlang-reg-send node (erlang-self) pid
                   (vector '$gen_cast (vector 'cast m f a (erlang-self)))))

;; rpc:call(node, module, function, arguments)
(define (erlang-rpc-call node m f . a)
  "Conducts a $gen_call to the RPC server."
  (apply erlang-gen-call node 'rex m f a))

;; rpc:cast(node, module, function, arguments)
(define (erlang-rpc-cast node m f . a)
  "Conducts a $gen_cast to the RPC server."
  (apply erlang-gen-cast node 'rex m f a))

(define (erlang-echo-test node data)
  (erlang-rpc-call node 'echo 'echo data))

;;------------------------------------------------------------------------------
;; Unit Tests
;;------------------------------------------------------------------------------

(require (planet "test.ss" ("schematics" "schemeunit.plt" 2 10)))

;; Abbreviations for testing.
(define eb->t erlang-binary->term)
(define t->eb term->erlang-binary)

(define (trip-it obj)
  (eb->t (t->eb obj)))

;; Can't really test communication in isolation, so we test that
;; functions for converting to/from the wire protocol are at least
;; internally consistent.
(define erlang-tests
  (test-suite
   "unit tests for erlang.ss"
   [test-equal? "integer"
                (trip-it 1) 1]
   [test-equal? "symbol"
                (trip-it 'foo) 'foo]
   [test-equal? "string 1"
                (trip-it "bar") "bar"]
   [test-equal? "string 1"
                (trip-it "this is a string") "this is a string"]
   [test-equal? "binary"
                (trip-it #"abcde") #"abcde"]
   [test-equal? "list of strings"
                (trip-it (list "a" "b" "c")) (list "a" "b" "c")]
   [test-equal? "vector"
                (trip-it #(bar baz)) #(bar baz)]
   [test-equal? "list 1"
                (trip-it '(1 foo ())) '(1 foo ())]
   [test-equal? "list 2"
                (trip-it '((#(1 2)) (#(1 2))))  '((#(1 2)) (#(1 2)))]
   [test-equal? "pid"
                (trip-it (vector erlang-tag 'erlang-pid 'someone@somehost 0 0 0))
                (vector erlang-tag 'erlang-pid 'someone@somehost 0 0 0)]
   [test-equal? "port"
                (trip-it (vector erlang-tag 'erlang-port 'someone@somehost 0 0))
                (vector erlang-tag 'erlang-port 'someone@somehost 0 0)]))

#|
(require (planet "text-ui.ss" ("schematics" "schemeunit.plt" 2 10)))

(define (run-tests)
  (test/text-ui erlang-tests))
|#
