;;; pack.ss --- Scheme implementation of Perl's pack/unpack.
;;; Copyleft (]) 2008 Eric Sessoms <nubgames@gmail.com>

#lang scheme

(provide network->integer
         integer->network
         unpack
         unpack-values
         pack
         pack-tests)

(define (network->integer bstr)
  "Convert a string of bytes in network byte-order to an integer."
  (case (bytes-length bstr)
    [(2 4 8)
     (integer-bytes->integer bstr #f #t)]
    [else
     (for/fold ([acc 0]) ([b (in-bytes bstr)])
       (+ (* acc 256) b))]))

(define (integer->network n b)
  "Convert an integer to a string of bytes in network byte-order."
  (case b
    [(1)
     (bytes n)]
    [(2 4 8)
     (integer->integer-bytes n b #f #t)]
    [else
     (let*-values ([(radix)   (for/first ([i (list 8 4 2)] #:when (> b i)) i)]
                   [(div mod) (quotient/remainder n (expt 256 radix))])
       (bytes-append (integer->network div (- b radix))
                     (integer->network mod radix)))]))

(define (unpack format in)
  (define (unpack1 format acc #:len (len #f))
    (if (null? format)
        (if (= 1 (length acc))
            (car acc)
            (reverse acc))
        (if (eq? #\/ (car format))
            (unpack1 (cdr format) (cdr acc) #:len (car acc))
            (let ([value
                   (case (car format)
                     [(#\a)      ; byte string
                      (read-bytes len in)]
                     [(#\A #\Z)  ; ascii string
                      (bytes->string/latin-1 (read-bytes len in))]
                     [(#\C)      ; unsigned char
                      (network->integer (read-bytes 1 in))]
                     [(#\n)      ; network byte-order short
                      (network->integer (read-bytes 2 in))]
                     [(#\N)      ; network byte-order long
                      (network->integer (read-bytes 4 in))])])
              (unpack1 (cdr format) (cons value acc))))))
  (if (bytes? in)
      (let* ([input  (open-input-bytes in)]
             [result (unpack format input)])
        (close-input-port input)
        result)
      (unpack1 (string->list format) empty)))

(define (unpack-values format in)
  (apply values (unpack format in)))

(define (pack format #:out (out #f) . args)
  (define (pack1 format args)
    ;; Don't check args because we want a mismatch to produce an error.
    (unless (null? format)
      (if (and (pair? (cdr format)) (eq? #\/ (cadr format)))
          (pack1 (cons (car format) (cddr format))
                 (cons ((if (bytes? (car args)) bytes-length string-length) (car args))
                       args))
          (let ([value
                 (case (car format)
                   [(#\a)      ; byte string
                    (car args)]
                   [(#\A #\Z)  ; ascii string
                    (string->bytes/latin-1 (car args))]
                   [(#\C)      ; unsigned char
                    (integer->network (car args) 1)]
                   [(#\n)      ; network byte-order short
                    (integer->network (car args) 2)]
                   [(#\N)      ; network byte-order long
                    (integer->network (car args) 4)])])
            (write-bytes value out)
            (pack1 (cdr format) (cdr args))))))
  (if (not out)
      (let ([output (open-output-bytes)])
        (keyword-apply pack (list '#:out) (list output) format args)
        (close-output-port output)
        (get-output-bytes output))
      (pack1 (string->list format) args)))

(require (planet "test.ss" ("schematics" "schemeunit.plt" 2 10)))

(define pack-tests
  (test-suite
   "unit tests for pack.ss"
   [test-equal? "unsigned char 1"
                (pack "C" 42) #"*"]
   [test-equal? "unsigned char 2"
                (unpack "C" (pack "C" 42)) 42]
   [test-equal? "network byte-order short 1"
                (pack "n" 256) #"\1\0"]
   [test-equal? "network byte-order short 2"
                (unpack "n" (pack "n" 1001)) 1001]
   [test-equal? "network byte-order long 1"
                (pack "N" 256) #"\0\0\1\0"]
   [test-equal? "network byte-order long 2"
                (unpack "N" (pack "N" 1002001)) 1002001]
   [test-equal? "ascii string 1"
                (pack "n/A" "hello, world!") #"\0\rhello, world!"]
   [test-equal? "ascii string 2"
                (unpack "n/A" (pack "n/A" "this is a string")) "this is a string"]
   [test-equal? "byte string"
                (unpack "N/a" (pack "N/a" #"abcde")) #"abcde"]))

#|
(require (planet "text-ui.ss" ("schematics" "schemeunit.plt" 2 10)))

(define (run-tests)
  (test/text-ui pack-tests))
|#
